"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""


class JemuInterrupts:
    _TYPE_STRING = "type"
    _INT = "interrupt"
    RESET = "reset"

    def __init__(self, jemu_connection):
        self._pin_interrupt_callbacks = []
        self._jemu_connection = jemu_connection
        self._jemu_connection.register(self.receive_packet)

    def on_interrupt(self, callback):
        self._pin_interrupt_callbacks += callback

    def receive_packet(self, jemu_packet):
        if jemu_packet[self._TYPE_STRING] == self._INT:
            for callback in self._pin_interrupt_callbacks:
                callback(jemu_packet)



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
import os
import json
from .jemu_button import JemuButton
from .jemu_mem_peripheral import JemuMemPeripheral
from .jemu_bma280 import JemuBMA280
from .jemu_sudo import JemuSudo
from .jemu_bq24160 import JemuBQ24160
from .jemu_bq27421 import JemuBQ27421


class JemuBspParser:
    _EXTERNAL_PERIPHERAL = "External"
    _NRF52_PERIPHERAL = "nrf52832"

    _bsp_json_path = None

    def __init__(self, bsp_json_path):
        self._bsp_json_path = bsp_json_path

    def get_components(self, jemu_connection):
        components_list = []
        mcu_list = []

        if not os.path.isfile(self._bsp_json_path):
            raise Exception(self._bsp_json_path + ' is not found')
        elif not os.access(self._bsp_json_path, os.R_OK):
            raise Exception(self._bsp_json_path + ' is not readable')
        else:
            with open(self._bsp_json_path) as bsp_json_file:
                bsp_json = json.load(bsp_json_file)

                for component in bsp_json["components"]:
                    component_id = component["id"]
                    component_obj = None

                    if ("config" in component) and ("generators" in component["config"]):
                        generators = component["config"]["generators"]
                    else:
                        generators = None

                    if component["class"] == "Button":
                        component_obj = JemuButton(jemu_connection, component_id)

                    if component["class"] == "BME280":
                        component_obj = JemuMemPeripheral(jemu_connection, component_id, self._EXTERNAL_PERIPHERAL, generators)
                    
                    if component["class"] == "BQ24160":
                        component_obj = JemuBQ24160(jemu_connection, component_id, self._EXTERNAL_PERIPHERAL, generators)
                        
                    if component["class"] == "BQ27421":
                        component_obj = JemuBQ27421(jemu_connection, component_id, self._EXTERNAL_PERIPHERAL, generators)
                    
                    if component["class"] == "BMA280":
                        component_obj = JemuBMA280(jemu_connection, component_id, self._EXTERNAL_PERIPHERAL, generators)
                        components_list.append({"obj": component_obj, "name": component["name"]})

                    if component_obj is not None:
                        components_list.append({"obj": component_obj, "name": component["name"]})
                        
                    if component["class"] == "MCU":
                        mcu_list.append(component["name"])
                        peripheral_obj = JemuSudo(jemu_connection, 39, "nrf52832")
                        components_list.append({"obj": peripheral_obj, "name": "SUDO"})
                    
            return components_list



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
import os
import os.path
import serial
import array
from time import sleep

from .common import TimeoutException


class WrongStateError(Exception):
    pass


class JemuUart(object):
    """
    Represents a UART device.
    """
    _uart_device = None
    _uart_device_path = None

    def __init__(self, uart_device_path, vlab):
        self._uart_device = None
        self._uart_device_path = uart_device_path
        self._buffer = b''
        self._vlab = vlab
        try:
            os.unlink(uart_device_path)
        except:
            pass

    def remove(self):
        if not os.path.exists(self._uart_device_path):
            return

        if not os.path.islink(self._uart_device_path):
            raise Exception(self._uart_device_path + ' not symbolic link')

        os.unlink(self._uart_device_path)

    def open(self):
        if not os.path.islink(self._uart_device_path):
            raise Exception(self._uart_device_path +
                            ' not found or not symbolic link')

        self._uart_device = serial.Serial(self._uart_device_path, timeout=0.2, rtscts=True, dsrdtr=True)

    def close(self):
        if self._uart_device:
            self._uart_device.close()

    def read_line(self, line_separator=b'\r\n'):
        """
        Reads a single line from UART.

        :param line_separator: Substring for separating lines. Usually b'\\\\r\\\\n' or b'\\\\n'

        :return: A byte string of data read from the device, including line_separator
        """
        while line_separator not in self._buffer:
            temp_data = self._uart_device.read(1024)
            self._buffer += temp_data
        line_length = self._buffer.find(line_separator) + len(line_separator)
        data = self._buffer[:line_length]
        self._buffer = self._buffer[line_length:]
        return data

    def read(self):
        """

        :return: data available on the UART device. Returns an empty bytes string if no data is available.
        """
        self._buffer += self._uart_device.read(1024)
        data = self._buffer
        self._buffer = b''
        return data

    def wait_until_uart_receives(self, data, timeout=None):
        """
        Blocks until specified data is received on UART.
        If the device is paused, this function will continue the execution and will pause the device when the data is ready or timeout occured.
        If the device is running, the function will not pause the device.

        :param data: Expected bytes string.
        :param timeout: Emulation time specified in milliseconds. jumper.vlab.TimeoutException is raised if timeout occurred before data was available in the buffer.
        :return: The data available on in the UART buffer when the specified data was received. Note that this can be more than the data provided in the data parameter.
        """
        original_state = self._vlab.get_state()

        if original_state == 'running':
            self._vlab.pause()
        elif original_state != 'paused':
            raise WrongStateError('state should be paused or running but was {}'.format(original_state))

        if timeout is not None:
            self._vlab.stop_after_ms(timeout)

        self._vlab.resume()

        while data not in self._buffer and self._vlab.get_state() == 'running':
            temp_data = self._uart_device.read(1024)
            self._buffer += temp_data
            sleep(0.01)

        self._buffer += self._uart_device.read()

        if timeout is not None:
            self._vlab.cancel_stop()

        if original_state == 'running':
            self._vlab.resume()
        elif original_state == 'paused':
            self._vlab.pause()

        if data in self._buffer:
            tmp = self._buffer
            self._buffer = b''
            return tmp

        raise TimeoutException("timeout while waiting for data")

    def write(self, data):
        """
        Writes data to UART

        :param data: Data to write
        """
        byte_arr = array.array('B', data)
        while len(byte_arr) > 0:
            byte_written = self._uart_device.write(byte_arr)
            byte_arr = byte_arr[byte_written:]




"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from queue import Queue


class FieldDescriptor:
    _RESPONSE_TIMEOUT = 10

    _TYPE_STRING = "type"
    _VALUE = "value"
    _PERIPHERAL_ID = "peripheral_id"
    _PERIPHERAL_TYPE = "peripheral_type"
    _FIELD_NAME = "field_name"
    _COMMAND = "command"
    _COMMAND_SET_VALUE = "set_value"
    _COMMAND_GET_VALUE = "get_value"

    def __init__(self, jemu_connection, id, field_name, peripheral_type):
        self._id = id
        self._jemu_connection = jemu_connection
        self._field_name = field_name
        self._peripheral_type = peripheral_type
        self._queue = Queue()
        self._jemu_connection.register(self._peripheral_receive_callback)

    def set(self, val):
        self._jemu_connection.send_json(self._set_value_json(val))

    def get(self):
        self._jemu_connection.send_json(self._get_value_json())
        return self._queue.get(timeout=self._RESPONSE_TIMEOUT)

    def _set_value_json(self, value):
        return {
            self._FIELD_NAME: self._field_name,
            self._TYPE_STRING: self._COMMAND,
            self._VALUE: value,
            self._PERIPHERAL_TYPE: self._peripheral_type,
            self._PERIPHERAL_ID: self._id,
            self._COMMAND: self._COMMAND_SET_VALUE
        }

    def _get_value_json(self):
        return {
            self._FIELD_NAME: self._field_name,
            self._TYPE_STRING: self._COMMAND,
            self._PERIPHERAL_TYPE: self._peripheral_type,
            self._PERIPHERAL_ID: self._id,
            self._COMMAND: self._COMMAND_GET_VALUE
        }

    def _peripheral_receive_callback(self, jemu_packet):
        if 'type' in jemu_packet and jemu_packet["type"] == "get_value_response":
            self._got_message = True
            self._queue.put(jemu_packet["value"])

    
class JemuPeripheral:

    def __init__(self, jemu_connection, id, peripheral_type, generators):
        self._peripheral_type = peripheral_type
        self._id = id
        self._jemu_connection = jemu_connection
        
        if generators is not None:
            for field in generators:
                descriptor = FieldDescriptor(jemu_connection, id, field, peripheral_type)
                setattr(self, field, descriptor)



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from .__version__ import __version__
from .vlab import Vlab
from .vlab_hci_device import VirtualHciDevice, BluezError



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from jemu_mem_peripheral import JemuMemPeripheral


class JemuBQ27421(JemuMemPeripheral):
    _COMMAND = "command"
    _TYPE_STRING = "type"
    _PERIPHERAL_ID = "peripheral_id"
    _INTERRUPT = "interrupts"
    _PERIPHERAL_TYPE = "peripheral_type"
    _COMMAND_SET_INTERRUPT = "set_interrupt"

    def __init__(self, jemu_connection, id, peripheral_type, generators):
        JemuMemPeripheral.__init__(self, jemu_connection, id, peripheral_type, generators)

    def _battery_interrupt_json(self):
        return {
            self._TYPE_STRING: self._COMMAND,
            self._PERIPHERAL_ID: self._id,
            self._COMMAND: self._COMMAND_SET_INTERRUPT,
            self._PERIPHERAL_TYPE: self._peripheral_type
        }

    def interrupt(self):
        """
        This function activates an interrupt in the bq27421 peripheral.
        On interrupt a 1 ms pulse is sent out over gpout pin.
        """
        self._jemu_connection.send_json(self._battery_interrupt_json())
